# w.c.s. - web application for online forms
# Copyright (C) 2005-2013  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import json

from qommon.ctl import Command, make_option

from wcs.formdef import FormDef
from wcs.workflows import Workflow
from wcs.wf.jump import JumpWorkflowStatusItem, jump_and_perform


class CmdTriggerJumps(Command):
    '''Triggers all "jump trigger" for a formdef, given host publisher context

    source.json file format:
        [
            {
                "data": { "info_for_wf_data": 42, ... },
                "select": { "form_number": 1, ... }
            },
            ...
        ]
    '''

    name = 'trigger-jumps'

    def __init__(self):
        Command.__init__(self, [
                make_option('--trigger', metavar='TRIGGER', action='store',
                            dest='trigger', help='(mandatory) trigger name'),
                make_option('--vhost', metavar='VHOST', action='store',
                            dest='vhost'),
                make_option('--workflow-id', metavar='WORKFLOW_ID', action='store',
                            dest='workflow_id'),
                make_option('--formdef-id', metavar='FORMDEF_ID', action='store',
                            dest='formdef_id'),
                make_option('--all-formdata', metavar='ALL_FORMDATA',
                    action='store_true', dest='all_formdata'),
                ])

    def execute(self, base_options, sub_options, args):

        if not sub_options.all_formdata and not args:
            exit('missing input json file(s) (or --all-formdata flag)')
        if not sub_options.vhost:
            exit('you must specificy --vhost')
        if not sub_options.trigger:
            exit('you must specificy --trigger')
        if sub_options.formdef_id and sub_options.workflow_id:
            exit('specify --workflow-id or --formdef-id')

        import publisher
        self.config.remove_option('main', 'error_log')
        publisher.WcsPublisher.configure(self.config, sub_options.extra)
        publisher = publisher.WcsPublisher.create_publisher()
        publisher.app_dir = os.path.join(publisher.app_dir, sub_options.vhost)
        publisher.set_config()

        trigger = sub_options.trigger
        if not sub_options.all_formdata:
            rows = list(get_rows(args))
        else:
            rows = '__all__'

        if sub_options.formdef_id:
            formdef = FormDef.get(id=sub_options.formdef_id, ignore_errors=True)
            if not formdef:
                exit('formdef-id does not exist')
            select_and_jump_formdata(formdef, trigger, rows)
        else:
            if sub_options.workflow_id:
                workflow = Workflow.get(id=sub_options.workflow_id, ignore_errors=True)
                if not workflow:
                    exit('workflow does not exist')
                workflows = [workflow]
            else:
                workflows = Workflow.select()
            for workflow in workflows:
                status_ids = list(get_status_ids_accepting_trigger(workflow, trigger))
                if status_ids:
                    for formdef in [f for f in FormDef.select() if f.workflow_id == workflow.id]:
                        select_and_jump_formdata(formdef, trigger, rows, status_ids)


CmdTriggerJumps.register()

def exit(message):
    print >> sys.stderr, message
    sys.exit(1)

def get_rows(args):
    for arg in args:
        for row in json.load(file(arg)):
            yield row

def get_status_ids_accepting_trigger(workflow, trigger):
    for status in workflow.possible_status:
        for item in status.items:
            if isinstance(item, JumpWorkflowStatusItem) and \
                item.trigger == trigger:
                yield 'wf-%s' % status.id, item.status
                break

def get_formdata_accepting_trigger(formdef, trigger, status_ids=None):
    if status_ids is None:
        workflow = formdef.get_workflow()
        status_ids = get_status_ids_accepting_trigger(workflow, trigger)
    formdata_ids = []

    data_class = formdef.data_class()
    for status_id, jump_to in status_ids:
        formdata_ids = data_class.get_ids_with_indexed_value('status', status_id)
        for formdata_id in formdata_ids:
            yield data_class.get(id=formdata_id), jump_to

def match_row(substitution_variables, row):
    select = row['select']
    for key, value in select.iteritems():
        if str(substitution_variables.get(key)) != str(value):
            return False
    return True

def select_and_jump_formdata(formdef, trigger, rows, status_ids=None):
    for formdata, jump_to in get_formdata_accepting_trigger(formdef, trigger, status_ids):
        if rows == '__all__':
            print 'formdata %s jumps to status %s' % (formdata, jump_to)
            jump_and_perform(formdata, jump_to)
        else:
            substitution_variables = formdata.get_substitution_variables()
            for row in rows:
                if match_row(substitution_variables, row):
                    print 'formdata %s jumps to status %s' % (formdata, jump_to)
                    jump_and_perform(formdata, jump_to, row.get('data'))
                    break # next formdata

