# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from quixote import get_publisher, get_request, get_response, get_session, redirect
from quixote.directory import Directory
from quixote.html import TemplateIO, htmltext

from wcs.fields import WidgetField

from qommon import template
from qommon import get_logger
from qommon.form import *
from wcs.anonylink import AnonymityLink
from qommon.strftime import strftime

from qommon.admin.texts import TextsDirectory

from qommon import errors


class FileDirectory(Directory):
    _q_exports = []

    def __init__(self, formdata, field_no):
        self.formdata = formdata
        self.field_no = field_no

    def _q_lookup(self, component):
        file = self.formdata.data[self.field_no]

        if component and component != file.base_filename:
            raise errors.TraversalError()

        response = get_response()
        if file.content_type:
            response.set_content_type(file.content_type)
        else:
            response.set_content_type('application/octet-stream')
        if file.charset:
            response.set_charset(file.charset)
        if file.base_filename:
            if file.content_type.startswith('image/'):
                response.set_header(
                        'content-disposition', 'inline; filename="%s"' % file.base_filename)
            else:
                response.set_header(
                        'content-disposition', 'attachment; filename="%s"' % file.base_filename)

        return file.get_file_pointer().read()



class FilesDirectory(Directory):
    def __init__(self, formdata):
        self.formdata = formdata

    def _q_lookup(self, component):
        return FileDirectory(self.formdata, component)


class FormStatusPage(Directory):
    _q_exports = ['', 'download', 'json']
    _q_extra_exports = []

    def html_top(self, title = None):
        template.html_top(title = title, default_org = _('Forms'))

    def __init__(self, formdef, filled, register_workflow_subdirs=True):
        get_publisher().substitutions.feed(filled)
        self.formdef = formdef
        self.filled = filled
        for q in self._q_extra_exports:
            if not q in self._q_exports:
                self._q_exports.append(q)

        if self.formdef.workflow and register_workflow_subdirs:
            for name, directory in self.filled.get_workflow_subdirectories():
                self._q_exports.append(name)
                setattr(self, name, directory)

    def check_auth(self):
        session = get_session()
        mine = False
        user = get_request().user
        if user:
            if user.anonymous:
                anonylink = AnonymityLink.select(
                    lambda x: x.name_identifier == session.name_identifier and
                              x.formdata_type == 'form' and
                              x.formdata_def_id == self.formdef.id)
                if len(anonylink) == 1:
                    mine = True
            elif self.filled.is_submitter(user):
                mine = True
        else:
            if get_session() and get_session().is_anonymous_submitter(self.filled):
                mine = True

        self.check_receiver()
        return mine

    def json(self):
        self.check_auth()
        return self.export_to_json()

    def receipt_message(self, mine=False):
        if self.formdef.workflow:
            workflow_message = self.filled.display_workflow_message()
            if workflow_message:
                r = TemplateIO(html=True)
                if workflow_message.startswith('<'):
                    r += htmltext(workflow_message)
                else:
                    r += htmltext('<p>%s</p>' % workflow_message)
                return r.getvalue()

        r = TemplateIO(html=True)
        # behaviour if workflow doesn't display any message
        if self.filled.receipt_time is not None:
            tm = misc.localstrftime(self.filled.receipt_time)
        else:
            tm = '???'

        r += htmltext('<div id="receipt-intro">')
        if self.formdef.only_allow_one:
            r += TextsDirectory.get_html_text('form-recorded-allow-one', vars={'date': tm})
        else:
            r += TextsDirectory.get_html_text('form-recorded',
                    vars={'date': tm, 'number': self.filled.get_display_id()})

        if mine:
            handling_role = self.filled.get_handling_role()
            if handling_role and handling_role.details:
                endpoint_status = self.formdef.workflow.get_endpoint_status()
                r += htmltext('<p>')
                if self.filled.status in [x.id for x in endpoint_status]:
                    r += _('Your case is handled by:')
                else:
                    r += _('Your case has been handled by:')
                r += htmltext('</p>')
                r += htmltext('<p id="receiver">')
                r += htmltext(handling_role.details.replace(str('\n'), str('<br />')))
                r += htmltext('</p>')

        r += htmltext('</div>')

        return r.getvalue()

    def _q_index(self):
        mine = self.check_auth()
        if get_request().is_json():
            return self.export_to_json()
        get_logger().info('form %s - id: %s - view' % (self.formdef.name, self.filled.id))

        self.html_top(self.formdef.name)

        r = TemplateIO(html=True)

        r += self.receipt_message(mine=mine)
        r += self.receipt()
        r += self.history()

        form = None

        session = get_session()
        user = get_request().user
        form = self.filled.get_workflow_form(user)
        if form:
            form.add_captcha()

        if form and form.is_submitted():
            if not form.has_errors():
                url = self.submit(form, comment_only = True)
            if not form.has_errors():
                if url is None:
                    url = get_request().get_frontoffice_url()
                response = get_response()
                response.set_status(303)
                response.headers[str('location')] = url
                response.content_type = 'text/plain'
                return "Your browser should redirect you"

        if form:
            r += form.render()

        r += self.form_status_buttons()
        return r.getvalue()

    def export_to_json(self):
        get_response().set_content_type('application/json')
        return self.filled.export_to_json()

    def form_status_buttons(self):
        if not get_response().iframe_mode:
            r = TemplateIO(html=True)
            r += htmltext('<div class="back-home-button">')
            r += htmltext('<a href="%s">%s</a>') % (get_publisher().get_root_url(), _('Back Home'))
            r += htmltext('</div>')
            return r.getvalue()

    def history(self):
        if not self.filled.evolution:
            return
        if not self.formdef.is_user_allowed_read_status_and_history(get_request().user, self.filled):
            return
        r = TemplateIO(html=True)
        r += htmltext('<div class="bo-block" id="evolution-log">')
        r += htmltext('<h2 class="foldable">%s</h2>') % _('Log')
        r += htmltext('<dl id="evolutions">')
        hidden = False
        for evo in self.filled.evolution:
            if evo.status:
                wf_status = self.filled.get_status(evo.status)
                if wf_status and not wf_status.is_visible(self.filled, get_request().user):
                    hidden = True
                else:
                    hidden = False
            if hidden:
                continue

            r += htmltext("<dt>%s") % misc.localstrftime(evo.time)
            if evo.who:
                evo_author = None
                if evo.who == '_submitter':
                    evo_author = _('Original Submitter')
                    if self.filled.is_submitter(get_request().user):
                        evo_author = get_request().user.display_name
                    elif self.filled.user_id:
                        try:
                            evo_author = _('Original Submitter (%s)') % \
                                         get_publisher().user_class.get(self.filled.user_id).display_name
                        except KeyError:
                            pass
                else:
                    try:
                        evo_author = evo.get_author_name()
                    except KeyError:
                        pass
                if evo_author:
                    r += htmltext(' <span class="user">%s</span>') % evo_author
            r += htmltext('</dt>')
            r += htmltext('<dd>')
            if evo.status:
                r += htmltext('<span class="status">%s</span> ') % self.filled.get_status_label(evo.status)
            if evo.comment:
                if evo.comment.startswith(str('#pre')):
                    r += htmltext('<div class="comment"><pre>%s</pre></div>') % evo.comment[4:]
                else:
                    r += htmltext('<div class="comment">')
                    r += htmltext('<p>')
                    r += htmltext('<p>') + htmltext('\n').join(
                            [(x or htmltext('</p><p>')) for x in evo.comment.splitlines()])
                    r += htmltext('</p>')
                    r += htmltext('</div>')
            for t in evo.display_parts():
                r += t
            r += htmltext('</dd>')
        r += htmltext('</dl>')
        r += htmltext('</div>') # .bo-block #evolution-log
        return r.getvalue()

    def check_receiver(self):
        session = get_session()
        if not session or not session.user:
            if not self.filled.formdef.is_user_allowed_read(None, self.filled):
                raise errors.AccessUnauthorizedError()
        user = get_request().user
        if self.filled.formdef is None:
            raise errors.AccessForbiddenError()
        if not self.filled.formdef.is_user_allowed_read(user, self.filled):
            raise errors.AccessForbiddenError()
        return user

    def receipt(self,
            always_include_user = False,
            show_status = True,
            form_url = ''):
        user = get_request().user
        if not always_include_user and get_request().user and \
                get_request().user.id == self.filled.user_id:
            user = None
        else:
            try:
                user = get_publisher().user_class.get(self.filled.user_id)
            except KeyError:
                user = None

        # this is custom code so it is possible to mark forms as anonyms, this
        # is done through the VoteAnonymity field, this is very specific but
        # isn't generalised yet into an useful extension mechanism, as it's not
        # clear at the moment what could be useful.
        for f in self.formdef.fields:
            if f.key == 'vote-anonymity':
                user = None
                break

        r = TemplateIO(html=True)
        r += htmltext('<div class="bo-block" id="summary">')
        r += htmltext('<h2 class="foldable">%s</h2>') % _('Summary')
        r += htmltext('<div class="dataview">')

        if user:
            r += htmltext('<div class="field"><span class="label">%s</span>') % _('User name')
            r += htmltext('<span class="value">%s</span></div>') % user.display_name

        on_page = False
        on_disabled_page = False
        for f in self.formdef.fields:

            if f.type == 'page':
                on_disabled_page = False
                if not f.is_visible(self.filled.data, self.formdef):
                    on_disabled_page = True

                form_field = False
                for f1 in self.formdef.fields[self.formdef.fields.index(f)+1:]:
                    if f1.key == 'page':
                        break
                    if isinstance(f1, WidgetField):
                        form_field = True
                        break
                if form_field is False:
                    on_disabled_page = True

            if on_disabled_page:
                continue


            if f.type == 'page':
                if on_page:
                    r += htmltext('</div>')
                r += htmltext('<div class="page">')
                r += htmltext('<h3>%s</h3>') % f.label
                on_page = True
                continue

            if not self.filled.data.has_key(f.id):
                continue

            if f.store_display_value and ('%s_display' % f.id) in self.filled.data:
                value = self.filled.data['%s_display' % f.id]
            else:
                value = self.filled.data[f.id]

            if value is None or value == '':
                continue

            if not hasattr(f, str('get_view_value')):
                continue

            r += htmltext('<div class="field"><span class="label">%s</span> ') % f.label
            r += htmltext('<div class="value">')
            s = f.get_view_value(value)
            s = s.replace(str('[download]'), str('%sdownload' % form_url))
            r += s
            r += htmltext('</div></div>')


        if on_page:
            r += htmltext('</div>')

        if show_status and self.formdef.is_user_allowed_read_status_and_history(
                        get_request().user, self.filled):
            wf_status = self.filled.get_visible_status()
            if wf_status:
                r += htmltext('<p><span class="label">%s</span> ') % _('Status')
                r += htmltext('<span class="value">%s</span></p>') % wf_status.name

        r += htmltext('</div>') # .dataview
        r += htmltext('</div>') # .bo-block
        return r.getvalue()


    def status(self):
        user = self.check_receiver()
        form = None

        try:
            form = self.filled.get_workflow_form(user)
        except:
            # XXX: probably because there are mixed forms, with and without
            # workflows
            form = Form()

        if form and form.is_submitted() and not form.has_errors():
            url = self.submit(form)
            if url is None:
                url = get_request().get_frontoffice_url()
            response = get_response()
            response.set_status(303)
            response.headers[str('location')] = url
            response.content_type = 'text/plain'
            return "Your browser should redirect you"

        get_logger().info('form %s - id: %s - view status' % (self.formdef.name, self.filled.id))
        self.html_top('%s - %s' % (self.formdef.name, self.filled.id))
        r = TemplateIO(html=True)
        if self.filled.receipt_time is not None:
            tm = misc.localstrftime(self.filled.receipt_time)
        else:
            tm = '???'
        r += htmltext("<p>")
        r += _('The form has been recorded on %(date)s with the number %(number)s.') % {
                'date': tm, 'number': self.filled.id}
        r += htmltext("</p>")

        if self.filled.anonymised:
            r += htmltext('<div class="infonotice">')
            r += htmltext(_('This form has been anonymised on %(date)s.')) % {
                    'date': self.filled.anonymised.strftime(misc.date_format())}
            r += htmltext('</div>')

        r += self.receipt(always_include_user = True)

        if self.formdef.workflow:
            r += htmltext(self.filled.display_workflow_message())

        r += self.history()

        if form:
            r += form.render()

        r += htmltext('<a href="..">%s</a>') % _('Back to Listing')
        return r.getvalue()

    def submit(self, form, comment_only = False):
        current_status = self.filled.status
        user = get_request().user
        next_url = self.filled.handle_workflow_form(user, form)
        if next_url:
            return next_url
        if form.has_errors():
            return
        if current_status != self.filled.status:
            get_logger().info('form %s - id: %s - status -> %s' % (
                        self.formdef.name, self.filled.id, self.filled.status))
            try:
                self.check_auth()
            except errors.AccessError:
                # the user no longer has access to the form; redirect to a
                # different page
                if 'backoffice/' in [x[0] for x in get_response().breadcrumb]:
                    user = get_request().user
                    if user and (user.is_admin or self.formdef.is_of_concern_for_user(user)):
                        # user has access to the formdef, redirect to the
                        # listing.
                        return '..'
                    else:
                        return get_publisher().get_backoffice_url()
                else:
                    return get_publisher().get_root_url()

    def download(self):
        self.check_receiver()
        try:
            fn = get_request().form['f']
            f = self.filled.data[fn]
        except (KeyError, ValueError):
            raise errors.TraversalError()

        file = self.filled.data[fn]
        if not hasattr(file, 'content_type'):
            raise errors.TraversalError()

        if hasattr(file, 'base_filename') and file.base_filename:
            return redirect('files/%s/%s' % (fn, file.base_filename))
        else:
            return redirect('files/%s/' % fn)

    def _q_lookup(self, component):
        if component == 'files':
            self.check_receiver()
            return FilesDirectory(self.filled)
        return Directory._q_lookup(self, component)

    def _q_traverse(self, path):
        get_response().breadcrumb.append(
            (str(self.filled.id) + '/',str(self.filled.id)))
        return super(FormStatusPage, self)._q_traverse(path)
